#include "File.h"
#include <stdio.h>
#include <iostream>
#include <sys/types.h>


#include "IsoPrimaryDescriptor.h"

void print_primary_descriptor(IsoPrimaryDescriptor *desc)
{
	printf("/-- * ISO Primary Descriptor * --\n|\n");
	printf("| System id: %s\n", desc->SystemID.c_str());
	printf("| Volume id: %s\n", desc->VolumeID.c_str());
	printf("| Number of sectors: %u\n", desc->NumberOfSectors);
	printf("| Logical block size: %u\n", (uint32_t)desc->SectorSize);
	uint64_t total_size = MAKE_LBA(desc->NumberOfSectors, desc->SectorSize);
	std::cout << "| Total volume size: " << total_size << " bytes = " << ((total_size/1024LL)/(1024LL)) << " MB" << std::endl;
	std::cout << "| Path table length: " << desc->PathTableSize << std::endl;
	std::cout << "| 1st directory record sector: " << desc->FirstPathTableSector << ", LBA = " << MAKE_LBA(desc->FirstPathTableSector, desc->SectorSize) << std::endl;
	std::cout << "| 2st directory record sector: " << desc->SecondPathTableSector << ", LBA = " << MAKE_LBA(desc->SecondPathTableSector, desc->SectorSize) << std::endl;
	std::cout << "| Volume set id: " << desc->VolumeID << std::endl;
	std::cout << "| Publisher id: " << desc->PublisherID << std::endl;
	std::cout << "| Preparer id: " << desc->PreparerID << std::endl;
	std::cout << "| Application id: " << desc->ApplicationID << std::endl;
	printf("|\n\\-- * End of Primary Descriptor * --\n\n");
}

void print_tree(DirectoryRecord *rec, int level)
{
	for (std::vector<DirectoryRecord*>::size_type i = 0; i < rec->children.size(); ++i)
	{
		for (int j = 0; j < level+1; ++j)
		{
			std::cout << " |";
		}
		
		if (rec->children[i]->isDirectory())
		{
			std::cout << "--" << rec->children[i]->Identifier << std::endl;
			print_tree(rec->children[i], level+1);
		}
		else
		{
			std::cout << "--"
				      << rec->children[i]->Identifier
					  << (rec->children[i]->isFinalRecord() ? "" : " *")
					  << std::endl;
		}
	}
}

bool readSectorIfNeeded(uint8_t *sector,
						TFile& dvd,
						IsoPrimaryDescriptor *desc,
						uint32_t& remaining,
						uint8_t*& current,
						uint32_t& sectors)
{
	if (remaining == 0 || *current == 0)
	{
		if (sectors == 0)
		{
			return false;
		}

		if (dvd.Read(sector, desc->SectorSize) != (ssize_t)desc->SectorSize)
		{
			return false;
		}

		--sectors;
		current = sector;
		remaining = desc->SectorSize;
	}

	return true;
}

void readSubdirectoryStructure(DirectoryRecord& parent,
							   TFile& dvd,
							   IsoPrimaryDescriptor* desc,
							   off_t lba,
							   uint32_t sectors,
							   uint8_t *sector)
{
	uint32_t remaining = 0;
	uint8_t *current = sector;

	if (dvd.SetPosition(lba) != lba)
	{
		return;
	}

	DirectoryRecord *r;

	// skip first 2 records
	for (int i = 0; i < 2; ++i)
	{
		if (!readSectorIfNeeded(sector, dvd, desc, remaining, current, sectors))
		{
			return;
		}
		r = DirectoryRecord::load(current, remaining);
		if (!r) return;
	}

	while (readSectorIfNeeded(sector, dvd, desc, remaining, current, sectors))
	{
		r = DirectoryRecord::load(current, remaining);
		if (!r) break;

		parent.children.push_back(r);
	}

	for (std::vector<DirectoryRecord*>::size_type i = 0; i < parent.children.size(); ++i)
	{
		if (parent.children[i]->isDirectory())
		{
			readSubdirectoryStructure(*parent.children[i], dvd, desc,
				MAKE_LBA(parent.children[i]->Sector, desc->SectorSize),
				parent.children[i]->Size / desc->SectorSize,
				sector);
		}
		else
		{
			// fix name
			std::string& name = parent.children[i]->Identifier;
			size_t p = name.find(';');
			if (p != std::string::npos)
			{
				name = name.substr(0, p);
			}
			if (name[name.length() - 1] == '.')
			{
				name = std::string(name, 0, name.length()-1);
			}
		}
	}
}


DirectoryRecord* readDirectoryStructure(TFile& dvd, IsoPrimaryDescriptor* desc, off_t lba)
{
	if (dvd.SetPosition(lba) != lba)
	{
		return 0;
	}

	uint8_t *sector = new uint8_t[desc->SectorSize];

	if (dvd.Read(sector, desc->SectorSize) != (ssize_t)desc->SectorSize)
	{
		delete []sector;
		return 0;
	}
	uint8_t *current = sector;
	uint32_t remaining = desc->SectorSize;

	DirectoryRecord *root_record = DirectoryRecord::load(current, remaining);
	if (!root_record) return 0;

	if (!root_record->isDirectory() || (root_record->Size % desc->SectorSize) != 0 || MAKE_LBA(root_record->Sector, desc->SectorSize) != lba)
	{
		delete []sector;
		delete root_record;
		return 0;
	}

	uint32_t sectors = root_record->Size / desc->SectorSize;
	
	readSubdirectoryStructure(*root_record, dvd, desc, lba, sectors, sector);
	delete []sector;
	return root_record;
}

DirectoryRecord* readDiskStructure(TFile& dvd, IsoPrimaryDescriptor* desc)
{
	off_t pos_req = MAKE_LBA(desc->FirstPathTableSector, desc->SectorSize);

	if(dvd.SetPosition(pos_req) != pos_req)
	{
		std::cout << "   ERROR: Seek to LBA = " << pos_req << " failed" << std::endl;
		return 0;
	}

	if (desc->PathTableSize < 10)
	{
		std::cout << "   ERROR: Primary descriptor does not contain root directory" << std::endl;
		return 0;
	}

	uint8_t *root_sector = new uint8_t[desc->SectorSize];
	
	if (dvd.Read(root_sector, desc->SectorSize) != (ssize_t)desc->SectorSize)
	{
		std::cout << "   ERROR: Cannot read path table" << std::endl;
		delete []root_sector;
		return 0;
	}

	uint8_t *current = root_sector;
	uint32_t remaining = desc->SectorSize;

	PathTableEntry *e = PathTableEntry::load(current, remaining);
	delete []root_sector;

	if (!e)
	{
		std::cout << "   ERROR: Cannot read path table" << std::endl;
		return 0;
	}

	off_t root_record = MAKE_LBA(e->FistDirectorySector, desc->SectorSize);

	return readDirectoryStructure(dvd, desc, root_record);
}

void print_primary_structure(DirectoryRecord *rec)
{
	if (rec)
	{
		std::cout << std::endl << " /ROOT" << std::endl;
		print_tree(rec, 0);
	}
}

bool parseCommand(const std::string& cmd, std::vector<std::string>& command)
{
	command.clear();

	std::string param;
	size_t index = 0;
	bool quote = false;

	char c;

	while (index < cmd.length())
	{
		c = cmd[index];
		if (c == ' ' && !quote)
		{
			if (param.length() > 0)
			{
				command.push_back(param);
				param.clear();
			}
		}
		else if (c == '\"')
		{
			quote = !quote;
		}
		else
		{
			param += c;
		}
		++index;
	}

	if (param.length() > 0)
	{
		command.push_back(param);
	}

	return command.size() > 0;
}

bool readCommand(std::vector<std::string>& command)
{
	static char cmd[2048];
	do
	{
		std::cout << " # ";
		std::cin.getline(cmd, 2047);
	} while (!parseCommand(cmd, command));

	if (command[0] == "exit")
	{
		return false;
	}
	return true;
}

void findFile(DirectoryRecord *rec, const std::string& src, std::vector<DirectoryRecord*>& out)
{
	size_t start = 0;
	size_t end = 0;
	
	std::vector<std::string> parts;
	std::string part;
	
	while (start < src.length())
	{
		end = src.find('/');
		if (end != (start + 1))
		{
			part = std::string(src, start, end-1);
			//std::cout << part << std::endl;
			parts.push_back(part);
		}
		start = end;
	}
	
	if (parts.size() == 0) return;
	
	// find parent directory record
	bool found = true;
	
	for (size_t i = 0; found && (i < parts.size() - 1); ++i)
	{
		found = false;
		
		for (size_t j = 0; j < rec->children.size(); ++j)
		{
			if (rec->children[j]->isDirectory() && rec->children[j]->Identifier == parts[i])
			{
				rec = rec->children[j];
				found = true;
				break;
			}
		}
	}
	
	if (!found)
	{
		return;
	}
	
	for (size_t i = 0; i < rec->children.size(); ++i)
	{
		if (rec->children[i]->isFile() && rec->children[i]->Identifier == part)
		{
			out.push_back(rec->children[i]);
			if (rec->children[i]->isFinalRecord()) break;
		}
	}
}

void cp(const std::string& src, const std::string& dest, TFile& dvd, IsoPrimaryDescriptor *desc, DirectoryRecord *rec)
{	
	std::vector<DirectoryRecord*> entries;
	findFile(rec, src, entries);
	
	if (entries.size() == 0)
	{
		std::cout << "File " << src << " was not found" << std::endl;
		return;
	}
	
	TFile output;
	
	if (!output.Open(dest.c_str(), TFile::Mode_Write | TFile::Mode_Create))
	{
		std::cout << "Cannot create output file" << std::endl;
		return;
	}
	
	std::cout << " * Copying file, this can take a long time ... " << std::endl;
	
	off_t lba;
	uint32_t size;
	uint32_t to_write;
	uint8_t *buffer = new uint8_t[desc->SectorSize];
	bool ok = true;
	
	for (size_t i = 0; i < entries.size() && ok; ++i)
	{
		lba = MAKE_LBA(entries[i]->Sector + entries[i]->ExtendedAttributeSize, desc->SectorSize);
		
		if (dvd.SetPosition(lba) != lba)
		{
			std::cout << " * ERROR: Source file is corrupted" << std::endl;
			ok = false;
			break;
		}
		
		size = entries[i]->Size;
		
		while (size > 0)
		{
			// bytes we need to write
			to_write = (size > (uint32_t)desc->SectorSize) ? (uint32_t)desc->SectorSize : size;
			
			// always read the whole sector
			if (dvd.Read(buffer, (size_t)desc->SectorSize) != (ssize_t)desc->SectorSize)
			{
				std::cout << " * ERROR: Source file is corrupted" << std::endl;
				ok = false;
				break;
			}
			
			// write bytes to output file
			if (output.Write(buffer, (size_t)to_write) != to_write)
			{
				std::cout << " * ERROR: Cannot write output file" << std::endl;
				ok = false;
				break;
			}
			
			size -= to_write;
		}
	}
	
	delete []buffer;
	output.Close();
	
	if (ok) std::cout << " * INFO: File succesfully copied" << std::endl;
}

void executeCommand(std::vector<std::string>& command,
					TFile& dvd,
					IsoPrimaryDescriptor *desc,
					DirectoryRecord *rec)
{
	if (command[0] == "help")
	{
		std::cout << "tree - show directory tree" << std::endl
		          << "cp - copy file from CD/DVD-ROM to local disk" << std::endl
		          << "exit - quit DVDReader" << std::endl;
	}
	else if (command[0] == "tree")
	{
		print_primary_structure(rec);
	}
	else if (command[0] == "cp")
	{
		if (command.size() < 3 || command.size() > 3)
		{
			std::cout << "Usage: cp file_on_cdrom file_on_disk" << std::endl;
		}
		else
		{
			cp(command[1], command[2], dvd, desc, rec);
		}
	}
	else
	{
		std::cout << "Unknown command \"" << command[0] << "\"" << std::endl;
	}
	std::cout << std::endl;

}

int main(int argc, char** argv)
{
	unsigned char sector[2048];
	IsoPrimaryDescriptor *primary_desc = 0;

	TFile f;

	char line[1024];
	std::string drive;
	
	std::cout << sizeof(off_t) << std::endl;

	std::cout << "Enter drive path: ";
	std::cin.getline(line, 1023);
	drive = line;

	if (!f.Open(drive.c_str(), TFile::Mode_Read))
	{
		std::cout << "Cannot open drive" << std::endl;
		return 0;
	}

	// skip first 16 sectors
	f.SetPosition(2048*16);

	bool do_read = true;
	printf("\n");

	while (do_read)
	{
		if (f.Read((void*)sector, 2048) != 2048)
		{
			return 0;
		}

		switch (sector[0])
		{
		case 0:
			printf(" * Ignoring boot record ...\n");
			break;
		case 1:
			if (primary_desc)
			{
				printf("WARNING: second ISO Primary Descriptor found, ignoring ...\n");
			}
			else
			{
				printf(" * Primary Descriptor found\n\n");
				primary_desc = new IsoPrimaryDescriptor();
				primary_desc->load(sector);
				print_primary_descriptor(primary_desc);
			}
			break;
		case 2:
			if (!primary_desc)
			{
				printf("WARNING: Supplementary Descriptor found but no Primary Descriptor exists.\n");
			}
			printf(" * Ignoring Supplementary Descriptor ...\n");
			break;
		case 255:
			printf(" * Terminating Descriptor found\n");
			do_read = false;
			break;
		default:
			printf(" * Unknown descriptor found, ignoring ...\n");
			break;
		}
	}

	if (!primary_desc)
	{
		std::cout << "ERROR: No primary descriptor." << std::endl;
		return 0;
	}

	std::cout << " * Reading path table for primary descriptor, this can take some time ..." << std::endl;
	DirectoryRecord *rec = readDiskStructure(f, primary_desc);
	
	if (!rec)
	{
		std::cout << "ERROR: Could NOT read path table." << std::endl;
		delete primary_desc;
		return 0;
	}

	std::vector<std::string> command;

	std::cout << " * Loading complete, command prompt active" << std::endl;

	std::cout << std::endl;

	while (readCommand(command))
	{
		executeCommand(command, f, primary_desc, rec);
	}

	delete rec;
	delete primary_desc;

	f.Close();

	return 0;
}
